import tempfile
import os
import os.path
import pathlib
import uuid
import json
import hashlib
import time
import random
import threading

from assemblyline.odm.models.config import Config, MetadataConfig
from assemblyline.remote.datatypes.queues.named import NamedQueue
from assemblyline_core.vacuum import crawler, worker
from assemblyline_core.ingester.constants import INGEST_QUEUE_NAME


def test_crawler(config: Config, redis_connection):
    try:
        with tempfile.TemporaryDirectory() as workdir:
            # Configure an environment
            data_dir = os.path.join(workdir, 'meta')
            os.mkdir(data_dir)

            config.core.vacuum.data_directories = [data_dir]
            config.core.vacuum.list_cache_directory = workdir
            config.core.vacuum.worker_cache_directory = workdir

            # Put some files named the right thing
            file_names = [uuid.uuid4().hex + '.meta' for _ in range(100)]
            for name in file_names:
                pathlib.Path(data_dir, name).touch()

            # Kick off a crawler thread
            threading.Thread(target=crawler.run, args=[config, redis_connection], daemon=True).start()

            # Check that they have been picked up
            queue = NamedQueue(crawler.VACUUM_BUFFER_NAME, redis_connection)
            while file_names:
                path = queue.pop(timeout=1)
                if path is None:
                    assert False
                file_names = [f for f in file_names if not path.endswith(f)]
    finally:
        # shut down the crawler
        crawler.stop_event.set()


def test_worker(config: Config, redis_connection):
    try:
        with tempfile.TemporaryDirectory() as workdir:
            # Configure an environment
            data_dir = os.path.join(workdir, 'meta')
            file_dir = os.path.join(workdir, 'files')
            os.mkdir(data_dir)
            os.mkdir(file_dir)

            config.core.vacuum.data_directories = [data_dir]
            config.core.vacuum.file_directories = [file_dir]
            config.core.vacuum.list_cache_directory = workdir
            config.core.vacuum.worker_cache_directory = workdir
            config.core.vacuum.assemblyline_user = 'service-account'
            config.core.vacuum.safelist = [{
                'name': 'good_streams',
                'conditions': {
                    'stream': '10+'
                }
            }]
            config.core.vacuum.worker_threads = 1

            # Apply strict metadata validation
            config.submission.metadata.ingest = {
                config.core.vacuum.ingest_type: {
                    'stream': {
                        'validator_type': 'integer',
                        'required': True
                    }
                }
            }
            config.submission.metadata.strict_schemes = [config.core.vacuum.ingest_type]

            # Place a file that will be safe
            first_file = os.path.join(data_dir, uuid.uuid4().hex + '.meta')
            with open(first_file, 'w') as temp:
                temp.write(json.dumps({
                    'sha256': '0'*64,
                    'metadata': {'stream': '100'}
                }))

            # Place a file that will be ingested
            test_file = random.randbytes(100)
            sha256 = hashlib.sha256(test_file).hexdigest()
            name = os.path.join(file_dir, sha256[0], sha256[1], sha256[2], sha256[3])
            os.makedirs(name)
            with open(os.path.join(name, sha256), 'bw') as temp:
                temp.write(test_file)

            second_file = os.path.join(data_dir, uuid.uuid4().hex + '.meta')
            with open(second_file, 'w') as temp:
                temp.write(json.dumps({
                    'sha256': sha256,
                    'metadata': {'stream': '99'}
                }))

            # Start the worker
            threading.Thread(target=worker.run, args=[config, redis_connection, redis_connection], daemon=True).start()

            # Tell the worker about the files
            queue = NamedQueue(crawler.VACUUM_BUFFER_NAME, redis_connection)
            queue.push(first_file)
            queue.push(second_file)

            # Get a message from for the ingested file
            ingest_queue = NamedQueue(INGEST_QUEUE_NAME, redis_connection)
            ingested = ingest_queue.pop(timeout=20)
            assert ingested is not None
            assert ingested['files'][0]['sha256'] == sha256
            assert ingest_queue.length() == 0

            # Make sure all the meta files have been consumed
            rounds = 0
            while os.path.exists(first_file) or os.path.exists(second_file) and rounds < 10:
                time.sleep(0.1)
                rounds += 1
            assert not os.path.exists(first_file)
            assert not os.path.exists(second_file)

    finally:
        worker.stop_event.set()



from unittest import mock
import pytest
import time

from assemblyline.datastore.helper import AssemblylineDatastore
from assemblyline.odm.models.submission import SubmissionParams
from assemblyline.odm.models.filescore import FileScore

from assemblyline_core.ingester.ingester import IngestTask, _dup_prefix, Ingester

from mocking import TrueCountTimes


@pytest.fixture
def submit_harness(clean_redis, clean_datastore: AssemblylineDatastore):
    """Setup a test environment just file for the ingest tests"""
    datastore = clean_datastore
    submitter = Ingester(datastore=datastore, redis=clean_redis, persistent_redis=clean_redis)
    submitter.running = TrueCountTimes(1)
    submitter.counter.increment = mock.MagicMock()
    submitter.submit_client.submit = mock.MagicMock()
    return datastore, submitter


def test_submit_simple(submit_harness):
    datastore, submitter = submit_harness

    # Push a normal ingest task
    submitter.unique_queue.push(0, IngestTask({
        'submission': {
            'params': SubmissionParams({
                'classification': 'U',
                'description': 'file abc',
                'services': {
                    'selected': [],
                    'excluded': [],
                    'resubmit': [],
                },
                'submitter': 'user',
            }),
            'files': [{
                'sha256': '0' * 64,
                'size': 100,
                'name': 'abc',
            }],
            'metadata': {}
        },
        'ingest_id': '123abc'
    }).as_primitives())
    submitter.handle_submit()

    # The task has been passed to the submit tool and there are no other submissions
    submitter.submit_client.submit.assert_called()
    assert submitter.unique_queue.pop() is None


def test_submit_duplicate(submit_harness):
    datastore, submitter = submit_harness

    # a normal ingest task
    task = IngestTask({
        'submission': {
            'params': SubmissionParams({
                'classification': 'U',
                'description': 'file abc',
                'services': {
                    'selected': [],
                    'excluded': [],
                    'resubmit': [],
                },
                'submitter': 'user',
            }),
            'files': [{
                'sha256': '0' * 64,
                'size': 100,
                'name': 'abc',
            }],
            'metadata': {}
        },
        'ingest_id': 'abc123'
    })
    # Make sure the scan key is correct, this is normally done on ingest
    task.submission.scan_key = task.params.create_filescore_key(task.submission.files[0].sha256, [])

    # Add this file to the scanning table, so it looks like it has already been submitted + ingest again
    submitter.scanning.add(task.submission.scan_key, task.as_primitives())
    submitter.unique_queue.push(0, task.as_primitives())

    submitter.handle_submit()

    # No tasks should be left in the queue
    assert submitter.unique_queue.pop() is None
    # The task should have been pushed to the duplicates queue
    assert submitter.duplicate_queue.length(_dup_prefix + task.submission.scan_key) == 1


def test_existing_score(submit_harness):
    datastore, submitter = submit_harness
    get_if_exists = datastore.filescore.get_if_exists
    try:
        # Set everything to have an existing filestore
        datastore.filescore.get_if_exists = mock.MagicMock(return_value=FileScore(
            dict(psid='000', expiry_ts=0, errors=0, score=10, sid='000', time=time.time())))

        # add task to internal queue
        submitter.unique_queue.push(0, IngestTask({
            'submission': {
                'params': SubmissionParams({
                    'classification': 'U',
                    'description': 'file abc',
                    'services': {
                        'selected': [],
                        'excluded': [],
                        'resubmit': [],
                    },
                    'submitter': 'user',
                }),
                'files': [{
                    'sha256': '0' * 64,
                    'size': 100,
                    'name': 'abc',
                }],
                'metadata': {},
                'notification': {
                    'queue': 'our_queue'
                }
            },
            'ingest_id': 'abc123'
        }).as_primitives())

        submitter.handle_submit()

        # No tasks should be left in the queue
        assert submitter.unique_queue.pop() is None
        # We should have received a notification about our task, since it was already 'done'
        assert submitter.notification_queues['nq-our_queue'].length() == 1
    finally:
        datastore.filescore.get_if_exists = get_if_exists



#!/usr/bin/env python
from __future__ import annotations

import concurrent.futures
import threading
import functools
import time
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, as_completed
from concurrent.futures.process import BrokenProcessPool
from typing import Callable, Optional, TYPE_CHECKING

import elasticapm
from datemath import dm

from assemblyline_core.server_base import ServerBase
from assemblyline_core.dispatching.dispatcher import BAD_SID_HASH
from assemblyline.common import forge
from assemblyline.common.isotime import epoch_to_iso, now_as_iso
from assemblyline.common.metrics import MetricsFactory
from assemblyline.filestore import FileStore
from assemblyline.odm.messages.expiry_heartbeat import Metrics
from assemblyline.remote.datatypes import get_client
from assemblyline.datastore.collection import Index
from assemblyline.remote.datatypes.set import Set

if TYPE_CHECKING:
    from assemblyline.datastore.collection import ESCollection


def file_delete_worker(logger, filestore_urls, file_batch, archive_filestore_urls=None) -> list[tuple[str, bool]]:
    try:
        filestore = FileStore(*filestore_urls)
        if archive_filestore_urls and filestore_urls != archive_filestore_urls:
            archivestore = FileStore(*archive_filestore_urls)
        else:
            archivestore = filestore

        def filestore_delete(item: tuple[str, bool]) -> tuple[Optional[str], Optional[bool]]:
            sha256, from_archive = item
            if from_archive:
                archivestore.delete(sha256)
                if not archivestore.exists(sha256):
                    return sha256, True
            else:
                filestore.delete(sha256)
                if not filestore.exists(sha256):
                    return sha256, False
            return None, None

        return _file_delete_worker(logger, filestore_delete, file_batch)

    except Exception as error:
        logger.exception("Error in filestore worker: " + str(error))
    return []


ActionSignature = Callable[[tuple[str, bool]], tuple[Optional[str], Optional[bool]]]


def _file_delete_worker(logger, delete_action: ActionSignature, file_batch) -> list[tuple[str, bool]]:
    finished_files: list[tuple[str, bool]] = []
    try:
        futures = []

        with ThreadPoolExecutor(8) as pool:
            for filename in file_batch:
                futures.append(pool.submit(delete_action, filename))

            for future in as_completed(futures):
                try:
                    erased_name, from_archive = future.result()
                    if erased_name and from_archive is not None:
                        finished_files.append((erased_name, from_archive))
                except Exception as error:
                    logger.exception("Error in filestore worker: " + str(error))

    except Exception as error:
        logger.exception("Error in filestore worker: " + str(error))
    return finished_files


class ExpiryManager(ServerBase):
    def __init__(self, redis_persist=None, datastore=None, filestore=None, config=None, classification=None):
        self.config = config or forge.get_config()

        super().__init__('assemblyline.expiry', shutdown_timeout=self.config.core.expiry.sleep_time + 5)

        # Set Archive related configs
        if self.config.datastore.archive.enabled:
            self.archive_access = True
            self.index_type = Index.HOT_AND_ARCHIVE
        else:
            self.archive_access = False
            self.index_type = Index.HOT

        self.datastore = datastore or forge.get_datastore(config=self.config, archive_access=self.archive_access)
        self.filestore = filestore or forge.get_filestore(config=self.config)
        self.classification = classification or forge.get_classification()
        self.expirable_collections: list[ESCollection] = []
        self.counter = MetricsFactory('expiry', Metrics)
        self.file_delete_worker = ProcessPoolExecutor(self.config.core.expiry.delete_workers)
        if self.config.filestore.archive:
            self.same_storage = self.config.filestore.storage == self.config.filestore.archive
        else:
            self.same_storage = True
        self.current_submission_cleanup = set()

        self.redis_persist = redis_persist or get_client(
            host=self.config.core.redis.persistent.host,
            port=self.config.core.redis.persistent.port,
            private=False,
        )
        self.redis_bad_sids = Set(BAD_SID_HASH, host=self.redis_persist)

        self.fs_hashmap = {
            'file': self.filestore_delete,
            'cached_file': self.cachestore_delete
        }

        for name, definition in self.datastore.ds.get_models().items():
            if hasattr(definition, 'expiry_ts'):
                self.expirable_collections.append(getattr(self.datastore, name))

        if self.config.core.metrics.apm_server.server_url is not None:
            self.log.info(f"Exporting application metrics to: {self.config.core.metrics.apm_server.server_url}")
            elasticapm.instrument()
            self.apm_client = forge.get_apm_client("expiry")
        else:
            self.apm_client = None

    @property
    def expiry_size(self):
        return self.config.core.expiry.delete_batch_size

    def stop(self):
        if self.counter:
            self.counter.stop()

        if self.apm_client:
            elasticapm.uninstrument()
        super().stop()

    def log_errors(self, function):
        @functools.wraps(function)
        def _func(*args, **kwargs):
            try:
                function(*args, **kwargs)
            except Exception:
                self.log.exception("Error in expiry worker")
        return _func

    def filestore_delete(self, file_batch, _):
        return self.file_delete_worker.submit(file_delete_worker, logger=self.log,
                                              filestore_urls=list(self.config.filestore.storage),
                                              file_batch=file_batch,
                                              archive_filestore_urls=list(self.config.filestore.archive))

    def cachestore_delete(self, file_batch, _):
        return self.file_delete_worker.submit(file_delete_worker, logger=self.log,
                                              filestore_urls=list(self.config.filestore.cache),
                                              file_batch=file_batch)

    def _finish_delete(self, collection: ESCollection, task: Future, expire_only: list[tuple[str, bool]]):
        # Wait until the worker process finishes deleting files
        file_list: list[str] = []
        while self.running:
            self.heartbeat()
            try:
                file_list = task.result(5)
                break
            except concurrent.futures.TimeoutError:
                pass

        if file_list:
            self.log.info(f'[{collection.name}] Deleted associated files from the '
                          f'{"cachestore" if "cache" in collection.name else "filestore"}...')
        else:
            self.log.info(f'[{collection.name}] Nothing was deleted from the '
                          f'{"cachestore" if "cache" in collection.name else "filestore"}...')

        # From the files to be deleted, check which are from the hot index
        hot_file_list = [x[0] for x in file_list if not x[1]]
        hot_file_list.extend([x[0] for x in expire_only if not x[1]])

        # From the files to be deleted, check which are from the archive index
        archive_file_list = [x[0] for x in file_list if x[1]]
        archive_file_list.extend([x[0] for x in expire_only if x[1]])

        for cur_file_list, index_type in [(hot_file_list, Index.HOT), (archive_file_list, Index.ARCHIVE)]:
            if not cur_file_list:
                # Nothing to delete from this index type
                continue

            # build a batch delete job for all the removed files
            bulk = collection.get_bulk_plan(index_type=index_type)
            for sha256 in cur_file_list:
                bulk.add_delete_operation(sha256)

            collection.bulk(bulk)
            self.counter.increment(f'{collection.name}', increment_by=len(cur_file_list))
            self.log.info(f"[{collection.name}] Deleted {len(cur_file_list)} items from the datastore...")

        if not hot_file_list and not archive_file_list:
            self.log.warning(f'[{collection.name}] Expiry unable to clean up any of the files in filestore.')

    def _simple_delete(self, collection: ESCollection, delete_query, number_to_delete):
        self.heartbeat()
        collection.delete_by_query(delete_query, index_type=self.index_type)
        self.counter.increment(f'{collection.name}', increment_by=number_to_delete)
        self.log.info(f"[{collection.name}] Deleted {number_to_delete} items from the datastore...")

    def _cleanup_canceled_submission(self, sid):
        # Allowing us at minimum 5 minutes to cleanup the submission
        self.heartbeat(int(time.time() + 5 * 60))
        if self.apm_client:
            self.apm_client.begin_transaction("Delete canceled submissions")

        # Cleaning up the submission
        self.log.info(f"[submission] Deleting incomplete submission {sid}...")
        self.datastore.delete_submission_tree_bulk(sid, self.classification, transport=self.filestore)
        self.redis_bad_sids.remove(sid)

        # We're done cleaning up the sid, mark it as done
        self.current_submission_cleanup.remove(sid)

        if self.apm_client:
            self.apm_client.end_transaction("canceled_submissions", 'deleted')

    def _process_chunk(self, collection: ESCollection, start, end, final_date, number_to_delete):
        # We assume that no records are ever inserted such that their expiry_ts is in the past.
        # We also assume that the `end` dates are also in the past.
        # As long as these two things are true, the set returned by this query should be consistent.
        # The one race condition is that a record might be refreshed while the file
        # blob would be deleted anyway, leaving a file record with no filestore object
        delete_query = f"expiry_ts:{{{start} TO {end}]"

        # check if we are dealing with an index that needs file cleanup
        if self.config.core.expiry.delete_storage and collection.name in self.fs_hashmap:
            # Delete associated files
            delete_objects: list[tuple[str, bool]] = []
            for item in collection.stream_search(
                    delete_query, fl='id,from_archive', as_obj=False, index_type=self.index_type):
                self.heartbeat()
                delete_objects.append((item['id'], item.get('from_archive', False)))

            # Filter archived documents if archive filestore is the same as the filestore
            expire_only: list[tuple[str, bool]] = []
            if self.same_storage and self.archive_access and collection.name == 'file':
                # Separate hot and archive files
                delete_from_archive = [i[0] for i in delete_objects if i[1]]
                delete_from_hot = [i[0] for i in delete_objects if not i[1]]

                # Check for overlap
                overlap = set(delete_from_archive).intersection(set(delete_from_hot))
                delete_from_archive = list(set(delete_from_archive)-overlap)
                delete_from_hot = list(set(delete_from_hot)-overlap)

                # Create the original delete_object form the overlap
                delete_objects = [(k, False) for k in overlap]
                delete_objects.extend([(k, True) for k in overlap])

                if delete_from_hot:
                    # Check hot objects to delete if they are in archive
                    archived_files = self.datastore.file.multiexists(delete_from_hot, index_type=Index.ARCHIVE)
                    delete_objects.extend([(k, False) for k, v in archived_files.items() if not v])
                    expire_only.extend([(k, False) for k, v in archived_files.items() if v])

                if delete_from_archive:
                    # Check hot objects to delete if they are in archive
                    hot_files = self.datastore.file.multiexists(delete_from_archive, index_type=Index.HOT)
                    delete_objects.extend([(k, True) for k, v in hot_files.items() if not v])
                    expire_only.extend([(k, True) for k, v in hot_files.items() if v])

            delete_tasks = self.fs_hashmap[collection.name](delete_objects, final_date)

            # Proceed with deletion, but only after all the scheduled deletes for this
            self.log.info(f"[{collection.name}] Scheduled {len(delete_objects)}/{number_to_delete} files to be "
                          f"removed from the {'cachestore' if 'cache' in collection.name else 'filestore'}")
            self._finish_delete(collection, delete_tasks, expire_only)

        else:
            # Proceed with deletion
            self._simple_delete(collection, delete_query, number_to_delete)

    def feed_expiry_jobs(self, collection, start, jobs: list[concurrent.futures.Future],
                         pool: ThreadPoolExecutor) -> tuple[str, bool]:
        _process_chunk = self.log_errors(self._process_chunk)
        number_to_delete = 0
        self.heartbeat()

        # Start of expiry transaction
        if self.apm_client:
            self.apm_client.begin_transaction("Delete expired documents")

        final_date = self._get_final_date()

        # Break down the expiry window into smaller chunks of data
        while len(jobs) < self.config.core.expiry.iteration_max_tasks:

            # Get the next chunk
            end, number_to_delete = self._get_next_chunk(collection, start, final_date)

            # Check if we got anything
            if number_to_delete == 0:
                break

            # Process the chunk in the threadpool
            jobs.append(pool.submit(_process_chunk, collection, start, end, final_date, number_to_delete))

            # Prepare for next chunk
            start = end

        # End of expiry transaction
        if self.apm_client:
            self.apm_client.end_transaction(collection.name, 'deleted')

        return start, number_to_delete < self.expiry_size

    def _get_final_date(self):
        now = now_as_iso()
        if self.config.core.expiry.batch_delete:
            final_date = dm(f"{now}||-{self.config.core.expiry.delay}h/d").float_timestamp
        else:
            final_date = dm(f"{now}||-{self.config.core.expiry.delay}h").float_timestamp
        return epoch_to_iso(final_date)

    def _get_next_chunk(self, collection: ESCollection, start, final_date):
        """Find date of item at chunk size and the number of items that
           will be affected in between start date and the date found"""
        rows = collection.search(f"expiry_ts: {{{start} TO {final_date}]", rows=1,
                                 offset=self.expiry_size - 1, sort='expiry_ts asc',
                                 as_obj=False, fl='expiry_ts', index_type=self.index_type)
        if rows['items']:
            return rows['items'][0]['expiry_ts'], self.expiry_size
        return final_date, rows['total']

    def try_run(self):
        pool = ThreadPoolExecutor(self.config.core.expiry.workers)
        main_threads = []

        # Launch a thread that will expire submissions that have been deleted
        thread = threading.Thread(target=self.clean_deleted_submissions, args=[pool])
        thread.start()
        main_threads.append(thread)

        # Launch threads that expire data from each collection of data
        for collection in self.expirable_collections:
            thread = threading.Thread(target=self.run_collection, args=[pool, collection])
            thread.start()
            main_threads.append(thread)

        # Wait for all the threads to exit
        for thread in main_threads:
            thread.join()

    def clean_deleted_submissions(self, pool):
        """Delete canceled submissions"""
        while self.running:
            # Make sure we're not dedicating more then a quarter of the pool to this operation because it is costly
            for submission in self.datastore.submission.search(
                    "to_be_deleted:true", fl="sid", rows=max(1, int(self.config.core.expiry.workers / 4)))['items']:
                if submission.sid not in self.current_submission_cleanup:
                    self.current_submission_cleanup.add(submission.sid)
                    pool.submit(self.log_errors(self._cleanup_canceled_submission), submission.sid)
            self.sleep_with_heartbeat(self.config.core.expiry.sleep_time)

    def run_collection(self, pool: concurrent.futures.ThreadPoolExecutor, collection):
        """Feed batches of jobs to delete to the thread pool for the given collection."""
        start = "*"
        jobs: list[concurrent.futures.Future] = []

        while self.running:
            try:
                try:
                    # Fill up 'jobs' with tasks that have been sent to the thread pool
                    # 'jobs' may already have items in it, but 'start' makes sure the new
                    # task added starts where the last finshed
                    start, final_job_small = self.feed_expiry_jobs(collection, start, jobs, pool)

                    # Wait until some of our work finishes and there is room in the queue for more work
                    finished, _jobs = concurrent.futures.wait(jobs, return_when=concurrent.futures.FIRST_COMPLETED)
                    jobs = list(_jobs)
                    for job in finished:
                        job.result()

                    # If we have expired all the data reset the start pointer
                    if len(jobs) == 0:
                        start = '*'

                except Exception as e:
                    self.log.exception(str(e))
                    continue

                # IF the most recent job added to the jobs list is short then
                # all the data is currently queued up to delete and we can sleep
                if final_job_small:
                    self.sleep_with_heartbeat(self.config.core.expiry.sleep_time)

            except BrokenProcessPool:
                self.log.error("File delete worker pool crashed.")
                self.file_delete_worker = ProcessPoolExecutor(self.config.core.expiry.delete_workers)


if __name__ == "__main__":
    with ExpiryManager() as em:
        em.serve_forever()



import logging

from assemblyline.common import forge
from assemblyline.common.isotime import iso_to_epoch, now_as_iso
from assemblyline.common.memory_zip import InMemoryZip
from assemblyline.datastore.helper import AssemblylineDatastore
from assemblyline.odm.messages.changes import Operation
from assemblyline.odm.models.service import SIGNATURE_DELIMITERS
from assemblyline.odm.models.signature import DEPLOYED_STATUSES, STALE_STATUSES, DRAFT_STATUSES


DEFAULT_DELIMITER = "\n\n"
CLASSIFICATION = forge.get_classification()


# Signature class
class SignatureClient:
    """A helper class to simplify signature management for privileged services and service-server."""

    def __init__(self, datastore: AssemblylineDatastore = None, config=None, classification_replace_map={}):
        self.log = logging.getLogger('assemblyline.signature_client')
        self.config = config or forge.CachedObject(forge.get_config)
        self.datastore = datastore or forge.get_datastore(self.config)
        self.service_list = forge.CachedObject(self.datastore.list_all_services, kwargs=dict(as_obj=False, full=True))
        self.delimiters = forge.CachedObject(self._get_signature_delimiters)
        self.classification_replace_map = classification_replace_map

    def _get_signature_delimiters(self):
        signature_delimiters = {}
        for service in self.service_list:
            if service.get("update_config", {}).get("generates_signatures", False):
                signature_delimiters[service['name'].lower()] = self._get_signature_delimiter(service['update_config'])
        return signature_delimiters

    def _get_signature_delimiter(self, update_config):
        delimiter_type = update_config['signature_delimiter']
        if delimiter_type == 'custom':
            delimiter = update_config['custom_delimiter'].encode().decode('unicode-escape')
        else:
            delimiter = SIGNATURE_DELIMITERS.get(delimiter_type, '\n\n')
        return {'type': delimiter_type, 'delimiter': delimiter}

    def _update_classification(self, signature):
        classification = signature['classification']
        # Update classification of signatures based on rewrite definition
        for term, replacement in self.classification_replace_map.items():
            if replacement.startswith('_'):
                # Replace with known field in Signature model
                # Otherwise replace with literal
                if signature.get(replacement[1:]):
                    replacement = signature[replacement[1:]]

            classification = classification.replace(term, replacement)

        # Save the (possibly) updated classfication
        signature['classification'] = classification


    def add_update(self, data, dedup_name=True):
        if data.get('type', None) is None or data['name'] is None or data['data'] is None:
            raise ValueError("Signature id, name, type and data are mandatory fields.")

        # Compute signature ID if missing
        data['signature_id'] = data.get('signature_id', data['name'])

        key = f"{data['type']}_{data['source']}_{data['signature_id']}"

        # Test signature name
        if dedup_name:
            check_name_query = f"name:\"{data['name']}\" " \
                f"AND type:\"{data['type']}\" " \
                f"AND source:\"{data['source']}\" " \
                f"AND NOT id:\"{key}\""
            other = self.datastore.signature.search(check_name_query, fl='id', rows='0')
            if other['total'] > 0:
                raise ValueError("A signature with that name already exists")

        old = self.datastore.signature.get(key, as_obj=False)
        op = Operation.Modified if old else Operation.Added
        if old:
            if old['data'] == data['data']:
                return True, key, None

            # Ensure that the last state change, if any, was made by a user and not a system account.
            user_modified_last_state = old['state_change_user'] not in ['update_service_account', None]

            # If rule state is moving to an active state but was disabled by a user before:
            # Keep original inactive state, a user changed the state for a reason
            if user_modified_last_state and data['status'] == 'DEPLOYED' and data['status'] != old['status']:
                data['status'] = old['status']

            # Preserve last state change
            data['state_change_date'] = old['state_change_date']
            data['state_change_user'] = old['state_change_user']

            # Preserve signature stats
            data['stats'] = old['stats']

        self._update_classification(data)

        # Save the signature
        success = self.datastore.signature.save(key, data)
        return success, key, op

    def add_update_many(self, source, sig_type, data, dedup_name=True):
        if source is None or sig_type is None or not isinstance(data, list):
            raise ValueError("Source, source type and data are mandatory fields.")

        # Test signature names
        names_map = {x['name']: f"{x['type']}_{x['source']}_{x.get('signature_id', x['name'])}" for x in data}

        skip_list = []
        if dedup_name:
            for item in self.datastore.signature.stream_search(f"type: \"{sig_type}\" AND source:\"{source}\"",
                                                               fl="id,name", as_obj=False, item_buffer_size=1000):
                lookup_id = names_map.get(item['name'], None)
                if lookup_id and lookup_id != item['id']:
                    skip_list.append(lookup_id)

            if skip_list:
                data = [
                    x for x in data
                    if f"{x['type']}_{x['source']}_{x.get('signature_id', x['name'])}" not in skip_list]

        old_data = self.datastore.signature.multiget(list(names_map.values()), as_dictionary=True, as_obj=False,
                                                     error_on_missing=False)

        plan = self.datastore.signature.get_bulk_plan()
        for rule in data:
            key = f"{rule['type']}_{rule['source']}_{rule.get('signature_id', rule['name'])}"
            if key in old_data:
                # Ensure that the last state change, if any, was made by a user and not a system account.
                user_modified_last_state = old_data[key]['state_change_user'] not in ['update_service_account', None]

                # If rule state is moving to an active state but was disabled by a user before:
                # Keep original inactive state, a user changed the state for a reason
                if user_modified_last_state and rule['status'] == 'DEPLOYED' and rule['status'] != old_data[key][
                        'status']:
                    rule['status'] = old_data[key]['status']

                # Preserve last state change
                rule['state_change_date'] = old_data[key]['state_change_date']
                rule['state_change_user'] = old_data[key]['state_change_user']

                # Preserve signature stats
                rule['stats'] = old_data[key]['stats']

            self._update_classification(rule)

            plan.add_upsert_operation(key, rule)

        if not plan.empty:
            res = self.datastore.signature.bulk(plan)
            return {"success": len(res['items']), "errors": res['errors'], "skipped": skip_list}

        return {"success": 0, "errors": [], "skipped": skip_list}

    def change_status(self, signature_id, status, user={}):
        possible_statuses = DEPLOYED_STATUSES + DRAFT_STATUSES
        if status not in possible_statuses:
            raise ValueError(f"You cannot apply the status {status} on yara rules.")

        data = self.datastore.signature.get(signature_id, as_obj=False)
        if data:
            if user and not CLASSIFICATION.is_accessible(user['classification'],
                                                         data.get('classification', CLASSIFICATION.UNRESTRICTED)):
                raise PermissionError("You are not allowed change status on this signature")

            if data['status'] in STALE_STATUSES and status not in DRAFT_STATUSES:
                raise ValueError(f"Only action available while signature in {data['status']} "
                                 f"status is to change signature to a DRAFT status. ({', '.join(DRAFT_STATUSES)})")

            if data['status'] in DEPLOYED_STATUSES and status in DRAFT_STATUSES:
                raise ValueError(f"You cannot change the status of signature {signature_id} from "
                                 f"{data['status']} to {status}.")

            today = now_as_iso()
            uname = user.get('uname')

            if status not in ['DISABLED', 'INVALID', 'TESTING']:
                query = f"status:{status} AND signature_id:{data['signature_id']} AND NOT id:{signature_id}"
                others_operations = [
                    ('SET', 'last_modified', today),
                    ('SET', 'state_change_date', today),
                    ('SET', 'state_change_user', uname),
                    ('SET', 'status', 'DISABLED')
                ]
                self.datastore.signature.update_by_query(query, others_operations)

            operations = [
                ('SET', 'last_modified', today),
                ('SET', 'state_change_date', today),
                ('SET', 'state_change_user', uname),
                ('SET', 'status', status)
            ]

            return self.datastore.signature.update(signature_id, operations), data
        raise FileNotFoundError(f"Signature not found. ({signature_id})")

    def download(self, query=None, access=None) -> bytes:
        if not query:
            query = "*"

        output_files = {}

        signature_list = sorted(
            self.datastore.signature.stream_search(
                query, fl="signature_id,type,source,data,order", access_control=access, as_obj=False,
                item_buffer_size=1000),
            key=lambda x: x['order'])

        for sig in signature_list:
            out_fname = f"{sig['type']}/{sig['source']}"
            if self.delimiters.get(sig['type'], {}).get('type', None) == 'file':
                out_fname = f"{out_fname}/{sig['signature_id']}"
            output_files.setdefault(out_fname, [])
            output_files[out_fname].append(sig['data'])

        output_zip = InMemoryZip()
        for fname, data in output_files.items():
            separator = self.delimiters.get(fname.split('/')[0], {}).get('delimiter', DEFAULT_DELIMITER)
            output_zip.append(fname, separator.join(data))

        return output_zip.read()

    def update_available(self, since='', sig_type='*'):
        since = since or '1970-01-01T00:00:00.000000Z'
        last_update = iso_to_epoch(since)
        last_modified = iso_to_epoch(self.datastore.get_signature_last_modified(sig_type))

        return last_modified > last_update




import hashlib
import random
import time
from copy import deepcopy

import pytest
from assemblyline.common.forge import get_classification
from assemblyline.common.isotime import iso_to_epoch
from assemblyline.odm.random_data import (
    create_badlists,
    create_users,
    wipe_badlist,
    wipe_users,
)
from assemblyline.odm.randomizer import get_random_hash
from assemblyline_core.badlist_client import BadlistClient, InvalidBadhash

CLASSIFICATION = get_classification()

add_hash_file = "10" + get_random_hash(62)
add_error_hash = "11" + get_random_hash(62)
update_hash = "12" + get_random_hash(62)
update_conflict_hash = "13" + get_random_hash(62)
source_hash = "14" + get_random_hash(62)

BAD_SOURCE = {
    "classification": CLASSIFICATION.UNRESTRICTED,
    "name": "BAD",
    "reason": [
        "2nd stage for implant BAD",
        "Used by actor BLAH!"
    ],
    "type": "external"}

BAD2_SOURCE = {
    "classification": CLASSIFICATION.UNRESTRICTED,
    "name": "BAD2",
    "reason": [
        "Use for phishing"
    ],
    "type": "external"}

ADMIN_SOURCE = {
    "classification": CLASSIFICATION.UNRESTRICTED,
    "name": "admin",
    "reason": [
        "It's denifitely bad",
    ],
    "type": "user"}

USER_SOURCE = {
    "classification": CLASSIFICATION.UNRESTRICTED,
    "name": "user",
    "reason": [
        "I just feel like it!",
        "I just feel like it!",
    ],
    "type": "user"}


@pytest.fixture(scope="module")
def client(datastore_connection):
    try:
        create_users(datastore_connection)
        create_badlists(datastore_connection)
        yield BadlistClient(datastore_connection)
    finally:
        wipe_users(datastore_connection)
        wipe_badlist(datastore_connection)


# noinspection PyUnusedLocal
def test_badlist_add_file(client):
    # Generate a random badlist
    sl_data = {
        'attribution': None,
        'hashes': {'md5': get_random_hash(32),
                   'sha1': get_random_hash(40),
                   'sha256': add_hash_file,
                   'ssdeep': None,
                   'tlsh': None},
        'file': {'name': ['file.txt'],
                 'size': random.randint(128, 4096),
                 'type': 'document/text'},
        'sources': [BAD_SOURCE, ADMIN_SOURCE],
        'type': 'file'
    }
    sl_data_original = deepcopy(sl_data)

    # Insert it and test return value
    qhash, op = client.add_update(sl_data)
    assert qhash == add_hash_file
    assert op == 'add'

    # Load inserted data from DB
    ds_sl = client.datastore.badlist.get(add_hash_file, as_obj=False)

    # Test dates
    added = ds_sl.pop('added', None)
    updated = ds_sl.pop('updated', None)

    # File item will live forever
    assert ds_sl.pop('expiry_ts', None) is None

    assert added == updated
    assert added is not None and updated is not None

    # Make sure tag is none
    tag = ds_sl.pop('tag', None)
    assert tag is None

    # Test classification
    classification = ds_sl.pop('classification', None)
    assert classification is not None

    # Test enabled
    enabled = ds_sl.pop('enabled', None)
    assert enabled

    # Normalize classification in sources
    for source in ds_sl['sources']:
        source['classification'] = CLASSIFICATION.normalize_classification(source['classification'])

    # Test rest
    assert ds_sl == sl_data_original


def test_badlist_add_tag(client):
    tag_type = 'network.static.ip'
    tag_value = '127.0.0.1'
    hashed_value = f"{tag_type}: {tag_value}".encode('utf8')
    expected_qhash = hashlib.sha256(hashed_value).hexdigest()

    # Generate a random badlist
    sl_data = {
        'attribution': {
            'actor': ["SOMEONE!"],
            'campaign': None,
            'category': None,
            'exploit': None,
            'implant': None,
            'family': None,
            'network': None
        },
        'dtl': 15,
        'hashes': {'sha256': expected_qhash},
        'tag': {'type': tag_type,
                'value': tag_value},
        'sources': [BAD_SOURCE, ADMIN_SOURCE],
        'type': 'tag'
    }
    sl_data_original = deepcopy(sl_data)

    # Insert it and test return value
    qhash, op = client.add_update(sl_data)
    assert qhash == expected_qhash
    assert op == 'add'

    # Load inserted data from DB
    ds_sl = client.datastore.badlist.get(expected_qhash, as_obj=False)

    # Test dates
    added = ds_sl.pop('added', None)
    updated = ds_sl.pop('updated', None)

    # Tag item will live up to a certain date
    assert ds_sl.pop('expiry_ts', None) is not None

    assert added == updated
    assert added is not None and updated is not None

    # Make sure file is None
    file = ds_sl.pop('file', {})
    assert file is None

    # Test classification
    classification = ds_sl.pop('classification', None)
    assert classification is not None

    # Test enabled
    enabled = ds_sl.pop('enabled', None)
    assert enabled

    # Test rest, dtl should not exist anymore
    sl_data_original.pop('dtl', None)

    # Normalize classification in sources
    for source in ds_sl['sources']:
        source['classification'] = CLASSIFICATION.normalize_classification(source['classification'])

    for hashtype in ['md5', 'sha1', 'ssdeep', 'tlsh']:
        ds_sl['hashes'].pop(hashtype, None)

    # Test rest
    assert ds_sl == sl_data_original


def test_badlist_add_invalid(client):
    # Generate a random badlist
    sl_data = {
        'hashes': {'sha256': add_error_hash},
        'sources': [USER_SOURCE],
        'type': 'file'}

    # Insert it and test return value
    with pytest.raises(ValueError) as conflict_exc:
        client.add_update(sl_data, user={"uname": "test"})

    assert 'for another user' in conflict_exc.value.args[0]


def test_badlist_update(client):
    # Generate a random badlist
    sl_data = {
        'attribution': {
            'actor': None,
            'campaign': None,
            'category': None,
            'exploit': None,
            'implant': ['BAD'],
            'family': None,
            'network': None},
        'hashes': {'md5': get_random_hash(32),
                   'sha1': get_random_hash(40),
                   'sha256': update_hash,
                   'ssdeep': None,
                   'tlsh': None},
        'file': {'name': [],
                 'size': random.randint(128, 4096),
                 'type': 'document/text'},
        'sources': [BAD_SOURCE],
        'type': 'file'
    }
    sl_data_original = deepcopy(sl_data)

    # Insert it and test return value
    qhash, op = client.add_update(sl_data)
    assert qhash == update_hash
    assert op == 'add'

    # Load inserted data from DB
    ds_sl = client.datastore.badlist.get(update_hash, as_obj=False)

    # Normalize classification in sources
    for source in ds_sl['sources']:
        source['classification'] = CLASSIFICATION.normalize_classification(source['classification'])

    # Test rest
    assert {k: v for k, v in ds_sl.items() if k in sl_data_original} == sl_data_original

    u_data = {
        'attribution': {'implant': ['TEST'], 'actor': ['TEST']},
        'hashes': {'sha256': update_hash, 'tlsh': 'faketlsh'},
        'sources': [USER_SOURCE],
        'type': 'file'
    }

    # Insert it and test return value
    qhash, op = client.add_update(u_data)
    assert qhash == update_hash
    assert op == 'update'

    # Load inserted data from DB
    ds_u = client.datastore.badlist.get(update_hash, as_obj=False)

    # Normalize classification in sources
    for source in ds_u['sources']:
        source['classification'] = CLASSIFICATION.normalize_classification(source['classification'])

    assert ds_u['added'] == ds_sl['added']
    assert iso_to_epoch(ds_u['updated']) > iso_to_epoch(ds_sl['updated'])
    assert len(ds_u['sources']) == 2
    assert USER_SOURCE in ds_u['sources']
    assert BAD_SOURCE in ds_u['sources']
    assert 'TEST' in ds_u['attribution']['implant']
    assert 'BAD' in ds_u['attribution']['implant']
    assert 'TEST' in ds_u['attribution']['actor']
    assert 'faketlsh' in ds_u['hashes']['tlsh']


def test_badlist_update_conflict(client):
    # Generate a random badlist
    sl_data = {'hashes': {'sha256': update_conflict_hash}, 'file': {}, 'sources': [ADMIN_SOURCE], 'type': 'file'}

    # Insert it and test return value
    qhash, op = client.add_update(sl_data)
    assert qhash == update_conflict_hash
    assert op == 'add'

    # Insert the same source with a different type
    sl_data['sources'][0]['type'] = 'external'
    with pytest.raises(InvalidBadhash) as conflict_exc:
        client.add_update(sl_data)

    assert 'has a type conflict:' in conflict_exc.value.args[0]

def test_badlist_tag_normalization(client):
    tag_type = 'network.static.uri'
    tag_value = 'https://BaD.com/About'

    normalized_value = 'https://bad.com/About'
    hashed_value = f"{tag_type}: {normalized_value}".encode('utf8')
    expected_qhash = hashlib.sha256(hashed_value).hexdigest()

    # Generate a random badlist
    sl_data = {
        'attribution': {
            'actor': ["SOMEONE!"],
            'campaign': None,
            'category': None,
            'exploit': None,
            'implant': None,
            'family': None,
            'network': None
        },
        'dtl': 15,
        'tag': {'type': tag_type,
                'value': tag_value},
        'sources': [BAD_SOURCE, ADMIN_SOURCE],
        'type': 'tag'
    }

    client.add_update(sl_data)

    # Assert that item got created with the expected ID from the normalized tag value
    assert client.datastore.badlist.exists(expected_qhash)
    time.sleep(1)

    # Assert that the tag exists in either format (within reason)
    assert client.exists_tags({tag_type: [tag_value]})
    assert client.exists_tags({tag_type: [normalized_value]})



import pytest

from assemblyline.odm.models.submission import Submission
from assemblyline.odm.models.config import Config, DEFAULT_CONFIG
from assemblyline.odm.models.service import Service
from assemblyline.odm.randomizer import random_model_obj

from assemblyline_core.dispatching.dispatcher import Scheduler
from assemblyline_core.server_base import get_service_stage_hash, ServiceStage


@pytest.fixture(scope='module')
def redis(redis_connection):
    redis_connection.flushdb()
    yield redis_connection
    redis_connection.flushdb()


def dummy_service(name, stage, category='static', accepts='', rejects=None, docid=None, extra_data=False, monitored_keys=()):
    return Service({
        'name': name,
        'stage': stage,
        'category': category,
        'accepts': accepts,
        'uses_temp_submission_data': extra_data,
        'uses_tags': extra_data,
        'rejects': rejects,
        'version': '0',
        'enabled': True,
        'timeout': 2,
        'monitored_keys': list(monitored_keys),
        'docker_config': {
            'image': 'somefakedockerimage:latest'
        }
    }, docid=docid)


# noinspection PyUnusedLocal,PyMethodMayBeStatic
class FakeDatastore:
    def __init__(self):
        self.service = self

    def stream_search(self, *args, **kwargs):
        return []

    def list_all_services(self, full=True):
        return {
            'extract': dummy_service(
                name='extract',
                stage='pre',
                accepts='archive/.*',
            ),
            'AnAV': dummy_service(
                name='AnAV',
                stage='core',
                category='av',
                accepts='.*',
            ),
            'cuckoo': dummy_service(
                name='cuckoo',
                stage='core',
                category='dynamic',
                accepts='document/.*|executable/.*',
            ),
            'polish': dummy_service(
                name='polish',
                stage='post',
                category='static',
                accepts='.*',
            ),
            'not_documents': dummy_service(
                name='not_documents',
                stage='post',
                category='static',
                accepts='.*',
                rejects='document/*',
            ),
            'Safelist': dummy_service(
                name='Safelist',
                stage='pre',
                category='static',
                accepts='.*',
            )
        }.values()


def submission(selected, excluded):
    sub = random_model_obj(Submission)
    sub.params.services.selected = selected
    sub.params.services.excluded = excluded
    return sub


@pytest.fixture
def scheduler(redis):
    config = Config(DEFAULT_CONFIG)
    config.services.stages = ['pre', 'core', 'post']
    stages = get_service_stage_hash(redis)
    ds = FakeDatastore()
    for service in ds.list_all_services():
        stages.set(service.name, ServiceStage.Running)
    return Scheduler(ds, config, redis)


def test_schedule_simple(scheduler):
    schedule = scheduler.build_schedule(submission(['static', 'av'], ['dynamic']), 'document/word')
    for a, b in zip(schedule, [['Safelist'], ['AnAV'], ['polish']]):
        assert set(a) == set(b)


def test_schedule_no_excludes(scheduler):
    schedule = scheduler.build_schedule(submission(['static', 'av', 'dynamic'], []), 'document/word')
    assert all(set(a) == set(b) for a, b in zip(schedule, [['Safelist'], ['AnAV', 'cuckoo'], ['polish']]))


def test_schedule_all_defaults_word(scheduler):
    schedule = scheduler.build_schedule(submission([], []), 'document/word')
    assert all(set(a) == set(b) for a, b in zip(schedule, [['Safelist'], ['AnAV', 'cuckoo'], ['polish']]))


def test_schedule_all_defaults_zip(scheduler):
    schedule = scheduler.build_schedule(submission([], []), 'archive/zip')
    assert all(set(a) == set(b)
               for a, b in zip(schedule, [['extract', 'Safelist'], ['AnAV'], ['polish', 'not_documents']]))


def test_schedule_service_safelist(scheduler):
    # Safelist service should still be scheduled
    schedule = scheduler.build_schedule(submission(["Safelist"], []), 'document/word', file_depth=0)
    for a, b in zip(schedule, [["Safelist"], [], []]):
        assert set(a) == set(b)

    # Safelist service should NOT still be scheduled because we're not enforcing Safelist service by default
    # and deep_scan and ignore_filtering is OFF for this submission
    sub = submission(["Safelist"], [])
    sub.params.deep_scan = False
    sub.params.ignore_filtering = False
    schedule = scheduler.build_schedule(sub, 'document/word', file_depth=1)
    for a, b in zip(schedule, [[], [], []]):
        assert set(a) == set(b)

    # Safelist service should be scheduled because we're enabling deep_scan
    sub.params.deep_scan = True
    sub.params.ignore_filtering = False
    schedule = scheduler.build_schedule(sub, 'document/word', file_depth=1)
    for a, b in zip(schedule, [["Safelist"], [], []]):
        assert set(a) == set(b)

    # Safelist service should be scheduled because we're enabling ignore_filtering
    sub.params.deep_scan = False
    sub.params.ignore_filtering = True
    schedule = scheduler.build_schedule(sub, 'document/word', file_depth=1)
    for a, b in zip(schedule, [["Safelist"], [], []]):
        assert set(a) == set(b)



from __future__ import annotations
from typing import Dict, Optional, Set, cast

import logging
import os
import re

from assemblyline.common.forge import CachedObject, get_classification
from assemblyline.datastore.helper import AssemblylineDatastore
from assemblyline.odm.models.config import Config
from assemblyline.odm.models.service import Service
from assemblyline.odm.models.submission import Submission
from assemblyline_core.server_base import get_service_stage_hash, ServiceStage


# If you are doing development and you want the system to route jobs ignoring the service setup/teardown
# set an environment variable SKIP_SERVICE_SETUP to true for all dispatcher containers
SKIP_SERVICE_SETUP = os.environ.get('SKIP_SERVICE_SETUP', 'false').lower() in ['true', '1']

Classification = get_classification()


class Scheduler:
    """This object encapsulates building the schedule for a given file type for a submission."""

    def __init__(self, datastore: AssemblylineDatastore, config: Config, redis):
        self.datastore = datastore
        self.config = config
        self._services: Dict[str, Service] = {}
        self.services = cast(Dict[str, Service], CachedObject(self._get_services))
        self.service_stage = get_service_stage_hash(redis)
        self.c12n_services: Dict[str, Set[str]] = {}

    def build_schedule(self, submission: Submission, file_type: str, file_depth: int = 0,
                       runtime_excluded: Optional[list[str]] = None,
                       submitter_c12n: Optional[str] = Classification.UNRESTRICTED) -> list[dict[str, Service]]:
        # Get the set of all services currently enabled on the system
        all_services = dict(self.services)

        # Retrieve a list of services that the classfication group is allowed to submit to
        if submitter_c12n is None:
            accessible = set(all_services.keys())
        else:
            accessible = self.get_accessible_services(submitter_c12n)

        # Load the selected and excluded services by category
        excluded = self.expand_categories(submission.params.services.excluded)
        runtime_excluded = self.expand_categories(runtime_excluded or [])
        if not submission.params.services.selected:
            selected = [s for s in all_services.keys()]
        else:
            selected = self.expand_categories(submission.params.services.selected)

        if submission.params.services.rescan:
            selected.extend(self.expand_categories(submission.params.services.rescan))

        # If we enable service safelisting, the Safelist service shouldn't run on extracted files unless:
        #   - We're enforcing use of the Safelist service (we always want to run the Safelist service)
        #   - We're running submission with Deep Scanning
        #   - We want to Ignore Filtering (perform as much unfiltered analysis as possible)
        if "Safelist" in selected and file_depth and self.config.services.safelist.enabled and \
                not self.config.services.safelist.enforce_safelist_service \
                and not (submission.params.deep_scan or submission.params.ignore_filtering):
            # Alter schedule to remove Safelist, if scheduled to run
            selected.remove("Safelist")

        # Add all selected, accepted, and not rejected services to the schedule
        schedule: list[dict[str, Service]] = [{} for _ in self.config.services.stages]
        services = list(set(selected).intersection(accessible) - set(excluded) - set(runtime_excluded))
        selected = []
        skipped = []
        for name in services:
            service = all_services.get(name, None)

            if not service:
                skipped.append(name)
                logging.warning(f"Service configuration not found: {name}")
                continue

            accepted = not service.accepts or re.match(service.accepts, file_type)
            rejected = bool(service.rejects) and re.match(service.rejects, file_type)

            if accepted and not rejected:
                schedule[self.stage_index(service.stage)][name] = service
                selected.append(name)
            else:
                skipped.append(name)

        return schedule

    def expand_categories(self, services: list[str]) -> list[str]:
        """Expands the names of service categories found in the list of services.

        Args:
            services (list): List of service category or service names.
        """
        if services is None:
            return []

        services = list(services)
        categories = self.categories()

        found_services = []
        seen_categories: set[str] = set()
        while services:
            name = services.pop()

            # If we found a new category mix in it's content
            if name in categories:
                if name not in seen_categories:
                    # Add all of the items in this group to the list of
                    # things that we need to evaluate, and mark this
                    # group as having been seen.
                    services.extend(categories[name])
                    seen_categories.add(name)
                continue

            # If it isn't a category, its a service
            found_services.append(name)

        # Use set to remove duplicates, set is more efficient in batches
        return list(set(found_services))

    def categories(self) -> Dict[str, list[str]]:
        all_categories: dict[str, list[str]] = {}
        for service in self.services.values():
            try:
                all_categories[service.category].append(service.name)
            except KeyError:
                all_categories[service.category] = [service.name]
        return all_categories

    def get_accessible_services(self, user_c12n: str) -> Set[str]:
        if not self.c12n_services.get(user_c12n):
            # Cache services that are accessible to a classification group
            self.c12n_services[user_c12n] = {_ for _, service in dict(self.services).items()
                                             if Classification.is_accessible(user_c12n, service.classification)}

        return self.c12n_services[user_c12n]

    def stage_index(self, stage):
        return self.config.services.stages.index(stage)

    def _get_services(self):
        old, self._services = self._services, {}
        stages = self.service_stage.items()
        services: list[Service] = self.datastore.list_all_services(full=True)
        for service in services:
            if service.enabled:
                # Determine if this is a service we would wait for the first update run for
                # Assume it is set to running so that in the case of a redis failure we fail
                # on the side of waiting for the update and processing more, rather than skipping
                wait_for = service.update_config and (service.update_config.wait_for_update and not SKIP_SERVICE_SETUP)
                # This is a service that we wait for, and is new, so check if it has finished its update setup
                if wait_for and service.name not in old:
                    if stages.get(service.name, ServiceStage.Running) == ServiceStage.Running:
                        self._services[service.name] = service
                else:
                    self._services[service.name] = service
        return self._services



import hashlib
import logging

import yaml
from assemblyline.common import forge
from assemblyline.common.isotime import now_as_iso
from assemblyline.datastore.helper import AssemblylineDatastore
from assemblyline.odm.models.user import ROLES
from assemblyline.remote.datatypes.lock import Lock

from assemblyline_core import normalize_hashlist_item

CLASSIFICATION = forge.get_classification()


class InvalidSafehash(Exception):
    pass


# Safelist class
class SafelistClient:
    """A helper class to simplify safelisting for privileged services and service-server."""

    def __init__(self, datastore: AssemblylineDatastore = None, config=None):
        self.log = logging.getLogger('assemblyline.safelist_client')
        self.config = config or forge.CachedObject(forge.get_config)
        self.datastore = datastore or forge.get_datastore(self.config)

    def _preprocess_object(self, data: dict):
        # Remove any null classification that might already be set
        if 'classification' in data and not data.get('classification'):
            data.pop('classification')

        # Set defaults
        data.setdefault('classification', CLASSIFICATION.UNRESTRICTED)
        data.setdefault('hashes', {})
        data.setdefault('expiry_ts', None)
        if data['type'] == 'tag':
            # Remove file related fields
            data.pop('file', None)
            data.pop('hashes', None)
            data.pop('signature', None)

            tag_data = data.get('tag', None)
            if tag_data is None or 'type' not in tag_data or 'value' not in tag_data:
                raise ValueError("Tag data not found")

            # Normalize tag data before further processing
            tag_data['value'] = normalize_hashlist_item(tag_data['type'], tag_data['value'])

            hashed_value = f"{tag_data['type']}: {tag_data['value']}".encode('utf8')
            data['hashes'] = {
                'sha256': hashlib.sha256(hashed_value).hexdigest()
            }

        elif data['type'] == 'signature':
            # Remove file related fields
            data.pop('file', None)
            data.pop('hashes', None)
            data.pop('tag', None)

            sig_data = data.get('signature', None)
            if sig_data is None or 'name' not in sig_data:
                raise ValueError("Signature data not found")

            hashed_value = f"signature: {sig_data['name']}".encode('utf8')
            data['hashes'] = {
                'sha256': hashlib.sha256(hashed_value).hexdigest()
            }

        elif data['type'] == 'file':
            data.pop('signature', None)
            data.pop('tag', None)
            data.setdefault('file', {})

        # Ensure expiry_ts is set on tag-related items
        dtl = data.pop('dtl', None) or self.config.core.expiry.safelisted_tag_dtl
        if dtl:
            data['expiry_ts'] = now_as_iso(dtl * 24 * 3600)

        # Set last updated
        data['added'] = data['updated'] = now_as_iso()

        # Find the best hash to use for the key
        for hash_key in ['sha256', 'sha1', 'md5', 'tlsh', 'ssdeep']:
            qhash = data['hashes'].get(hash_key, None)
            if qhash:
                break

        # Validate hash length
        if not qhash:
            raise ValueError("No valid hash found")

        return qhash

    def add_update(self, safelist_object: dict, user: dict = None):
        qhash = self._preprocess_object(safelist_object)

        # Validate sources
        src_map = {}
        for src in safelist_object['sources']:
            if user:
                if src['type'] == 'user':
                    if src['name'] != user['uname']:
                        raise ValueError(f"You cannot add a source for another user. {src['name']} != {user['uname']}")
                else:
                    if ROLES.signature_import not in user['roles']:
                        raise PermissionError("You do not have sufficient priviledges to add an external source.")

            # Find the highest classification of all sources
            safelist_object['classification'] = CLASSIFICATION.max_classification(
                safelist_object['classification'], src.get('classification', None))

            src_map[src['name']] = src

        with Lock(f'add_or_update-safelist-{qhash}', 30):
            old = self.datastore.safelist.get_if_exists(qhash, as_obj=False)
            if old:
                # Save data to the DB
                self.datastore.safelist.save(qhash, SafelistClient._merge_hashes(safelist_object, old))
                return qhash, "update"
            else:
                try:
                    safelist_object['sources'] = list(src_map.values())
                    self.datastore.safelist.save(qhash, safelist_object)
                    return qhash, "add"
                except Exception as e:
                    return ValueError(f"Invalid data provided: {str(e)}")

    def add_update_many(self, list_of_safelist_objects: list):
        if not isinstance(list_of_safelist_objects, list):
            raise ValueError("Could not get the list of hashes")

        new_data = {}
        for safelist_object in list_of_safelist_objects:
            qhash = self._preprocess_object(safelist_object)
            new_data[qhash] = safelist_object

        # Get already existing hashes
        old_data = self.datastore.safelist.multiget(list(new_data.keys()), as_dictionary=True, as_obj=False,
                                                    error_on_missing=False)

        # Test signature names
        plan = self.datastore.safelist.get_bulk_plan()
        for key, val in new_data.items():
            # Use maximum classification
            old_val = old_data.get(key, {'classification': CLASSIFICATION.UNRESTRICTED,
                                         'hashes': {}, 'sources': [], 'type': val['type']})

            # Add upsert operation
            plan.add_upsert_operation(key, SafelistClient._merge_hashes(val, old_val))

        if not plan.empty:
            # Execute plan
            res = self.datastore.safelist.bulk(plan)
            return {"success": len(res['items']), "errors": res['errors']}

        return {"success": 0, "errors": []}

    def exists(self, qhash):
        return self.datastore.safelist.get_if_exists(qhash, as_obj=False)

    def get_safelisted_tags(self, tag_types):
        if isinstance(tag_types, str):
            tag_types = tag_types.split(',')

        with forge.get_cachestore('system', config=self.config, datastore=self.datastore) as cache:
            tag_safelist_yml = cache.get('tag_safelist_yml')
            if tag_safelist_yml:
                tag_safelist_data = yaml.safe_load(tag_safelist_yml)
            else:
                tag_safelist_data = forge.get_tag_safelist_data()

        if tag_types:
            output = {
                'match': {k: v for k, v in tag_safelist_data.get('match', {}).items()
                          if k in tag_types or tag_types == []},
                'regex': {k: v for k, v in tag_safelist_data.get('regex', {}).items()
                          if k in tag_types or tag_types == []},
            }
            for tag in tag_types:
                for sl in self.datastore.safelist.stream_search(
                        f"type:tag AND enabled:true AND tag.type:{tag}", as_obj=False):
                    output['match'].setdefault(sl['tag']['type'], [])
                    output['match'][sl['tag']['type']].append(sl['tag']['value'])

        else:
            output = tag_safelist_data
            for sl in self.datastore.safelist.stream_search("type:tag AND enabled:true", as_obj=False):
                output['match'].setdefault(sl['tag']['type'], [])
                output['match'][sl['tag']['type']].append(sl['tag']['value'])

        return output

    def get_safelisted_signatures(self, **_):
        output = [
            item['signature']['name']
            for item in self.datastore.safelist.stream_search(
                "type:signature AND enabled:true", fl="signature.name", as_obj=False)]

        return output

    @staticmethod
    def _merge_hashes(new, old):
        # Account for the possibility of merging with null types
        if not (new or old):
            # Both are null
            raise ValueError("New and old are both null")
        elif not (new and old):
            # Only one is null, in which case return the other
            return new or old

        try:
            # Check if hash types match
            if new['type'] != old['type']:
                raise InvalidSafehash(f"Safe hash type mismatch: {new['type']} != {old['type']}")

            # Use the new classification but we will recompute it later anyway
            old['classification'] = new['classification']

            # Update updated time
            old['updated'] = new.get('updated', now_as_iso())

            # Update hashes
            old['hashes'].update({k: v for k, v in new['hashes'].items() if v})

            # Update type specific info
            if old['type'] == 'file':
                old.setdefault('file', {})
                new_names = new.get('file', {}).pop('name', [])
                if 'name' in old['file']:
                    for name in new_names:
                        if name not in old['file']['name']:
                            old['file']['name'].append(name)
                elif new_names:
                    old['file']['name'] = new_names
                old['file'].update({k: v for k, v in new.get('file', {}).items() if v})
            elif old['type'] == 'tag':
                old['tag'] = new['tag']

            # Merge sources
            src_map = {x['name']: x for x in new['sources']}
            if not src_map:
                raise InvalidSafehash("No valid source found")

            old_src_map = {x['name']: x for x in old['sources']}
            for name, src in src_map.items():
                if name not in old_src_map:
                    old_src_map[name] = src
                else:
                    old_src = old_src_map[name]
                    if old_src['type'] != src['type']:
                        raise InvalidSafehash(f"Source {name} has a type conflict: {old_src['type']} != {src['type']}")

                    for reason in src['reason']:
                        if reason not in old_src['reason']:
                            old_src['reason'].append(reason)
                    old_src['classification'] = src.get('classification', old_src['classification'])
            old['sources'] = list(old_src_map.values())

            # Calculate the new classification
            for src in old['sources']:
                old['classification'] = CLASSIFICATION.max_classification(
                    old['classification'], src.get('classification', None))

            # Set the expiry
            old['expiry_ts'] = new.get('expiry_ts', None)
            return old
        except Exception as e:
            raise InvalidSafehash(f"Invalid data provided: {str(e)}")



PAUSABLE_COMPONENTS = ['ingester', 'dispatcher']

def normalize_hashlist_item(tag_type: str, tag_value: str) -> str:
    # Normalize tag data pertaining to domains or URIs
    if tag_type.endswith('.domain'):
        tag_value = tag_value.lower()
    elif tag_type.endswith('.uri'):
        hostname = tag_value.split('//', 1)[1].split('/', 1)[0]
        tag_value = tag_value.replace(hostname, hostname.lower(), 1)
    return tag_value
